
; Tell the AS assembler we want to code for NGPC
;
        cpu 96C141
        maxmode on
        include "stddef96.inc"


        include "hardware.inc"  ; Include hardware defines
        include "system.inc"    ; Include System Call defines
        
;
; Define variables
;


START_OF_RAM    EVAL    _MAINRAM
   RAMDB YHEAD
   RAMDB XHEAD
   RAMDB YTAIL
   RAMDB XTAIL
   RAMDB YFRT
   RAMDB XFRT
   RAMDB DIR
   RAMDB QHEAD
   RAMDB QTAIL
   RAMDB RANDOMOFS
   RAMDB OLDYHEAD
   RAMDB OLDXHEAD
   RAMDB OLDDIR
   RAMDB OLDYTAIL
   RAMDB OLDXTAIL
   RAMDB SCOREL
   RAMDB SCOREH
   RAMDB HIL
   RAMDB HIH
   RAMDB VBCOUNTER
   RAMDB SPEED

QBASE       EQU   4200h
FRUITCHAR   EQU   65h
HEADBASE    EQU   66h
TEXTBASE    EQU   49H

HEADER_OFS      EQU     0200000H

;
; Standard cartridge header
;
    org    HEADER_OFS
    db     " LICENSED BY SNK CORPORATION"    ; 28 bytes license string
    dd     Start                             ; Program Counter
    dw     0                                 ; Catalog number
    db     0                                 ; Sub catalog number
    db     10h                               ; colour or b+w (10h = colour)
    db     " SNAKE 1.2  "                    ; Game name (12 bytes)
    dd     0,0,0,0                           ; padding - reserve for future use

; User Interrupt Vectors

UserIntVect:
        dd      nada            ; Software Interrupt (SWI 3)
        dd      nada            ; Software Interrupt (SWI 4)
        dd      nada            ; Software Interrupt (SWI 5)
        dd      nada            ; Software Interrupt (SWI 6)
        dd      nada            ; RTC Alarm Interrupt
        dd      VBlankInt       ; Vertical Blanking Interrupt
        dd      nada            ; Interrupt from Z80
        dd      nada            ; Timer Interrupt (8 bit timer 0)
        dd      nada            ; Timer Interrupt (8 bit timer 1)
        dd      nada            ; Timer Interrupt (8 bit timer 2)
        dd      nada            ; Timer Interrupt (8 bit timer 3)
        dd      nada            ; Serial Transmission Interrupt
        dd      nada            ; Serial Reception Interrupt
        dd      nada            ; (Reserved)
        dd      nada            ; End Micro DMA Int (MicroDMA 0)
        dd      nada            ; End Micro DMA Int (MicroDMA 1)
        dd      nada            ; End Micro DMA Int (MicroDMA 2)
        dd      nada            ; End Micro DMA Int (MicroDMA 3)

nada:   reti

; *** Start of User Code ***

Start:
        calr    OS_VERSION      ; Initialize NGP or NGPC mode

        set     6,(rUSERA)      ; User Answer

; Install User Interrupt Vectors

        lda     xix,(UserIntVect)
        lda     xiy,(rSWI3)
        ld      b,18
UIVloop:
        ld      xwa,(xix+)
        ld      (xiy+),xwa
        djnz    b,UIVloop


    ; set up screen size 160x152
    ld   (08002h),0
    ld   (08003h),0
    ld   (08004h),0a0h
    ld   (08005h),98h

    ; clear hiscore
    ld   (HIL), 0
    ld   (HIH), 0
    
    ; copy charset to character RAM
    ld bc,(ENDCHARSET-CHARSET)/2
    ld xde, _TILERAM+400h
    ld xhl, CHARSET
    ldirw (xde+),(xhl+)

; Enable Interrupts

        ei      0

Restart:
    ; copy palette to character palette RAM
    ld bc, (ENDPALETTE-PALETTE)/2
    ld xde, _SCR2PAL
    ld xhl, PALETTE
    ldirw (xde+),(xhl+)

    ; background colour 
    ; _BGCPAL = GRxB
    ld (_BGCPAL), 36h
    ld (_BGCPAL+1), 00h

    ld (SPEED), 8 ; initially slow speed

    call TitleScreen
    call ClearScreen

    call qinit
    call varinit

    ; draw fruit
    ld hl, (YFRT)
    ld w, 0
    ld a, FRUITCHAR
    call DrawChar
    
MAIN
    ; remember position of previous head
    ld wa, (YHEAD)
    ld (OLDYHEAD), wa
    ld a, (DIR)
    ld (OLDDIR), a
    ld wa, (YTAIL)
    ld (OLDYTAIL), wa

    ld a, (rSL)
    call ValidControl

l1
    ; check up
    bit PADB_UP, a
    jr z, ldown
    ld (DIR), PADF_UP
    ld c, (YHEAD)
    cp c, 1
    jr nz, lu1
    ld c, 19
lu1 sub c, 1
    ld (YHEAD), c
    jp postcontrol

ldown
    ; check down
    bit PADB_DOWN, a
    jr z, lleft
    ld (DIR), PADF_DOWN
    ld c, (YHEAD)
    cp c, 18
    jr nz, ld1
    ld c, 0
ld1 add c, 1
    ld (YHEAD), c
    jp postcontrol

lleft
    ; check left
    bit PADB_LEFT, a
    jr z, lright
    ld (DIR), PADF_LEFT
    ld c, (XHEAD)
    cp c, 0
    jr nz, ll1
    ld c, 20
ll1 sub c, 1
    ld (XHEAD), c
    jp postcontrol

lright
    ; check right
    bit PADB_RIGHT, a
    jr z, postcontrol
    ld (DIR), PADF_RIGHT
    ld c, (XHEAD)
    cp c, 19
    jr nz, lr1
    ld c, 0
    jr lr2
lr1 add c, 1
lr2 ld (XHEAD), c

postcontrol
    ; store head direction
    ld l, (DIR)
    call qadd
 
    call CheckDeath
    cp a, 0
    jr z, cdb
cda call UpdateHighScore
    jp Restart
cdb
    ; draw head
    ld xbc, WHICHHEAD
    ld xhl, 0
    ld l, (DIR)
    add xbc, xhl
    ld w, 2
    ld a, HEADBASE
    add a, (xbc)
    ld hl, (YHEAD)
    call DrawChar

    ; blank old head with body in correct direction
    ld a, 0
    ld xbc, DIRCONV
    ld xhl, 0
    ld l, (DIR)
    add xbc, xhl
    add a, (xbc)
    ld xbc, ODIRCONV
    ld xhl, 0
    ld l, (OLDDIR)
    add xbc, xhl
    add a, (xbc)
    ld xhl, 0
    ld l, a
    ld xbc, BODYTABLE
    add xbc, xhl
    ld a, HEADBASE
    add a, (xbc)
    ld w, 2
    ld hl, (OLDYHEAD)
    call DrawChar     


    ; check if head = fruit
    ld l, (XHEAD)
    cp l, (XFRT)
    jr nz, nothit
    ld l, (YHEAD)
    cp l, (YFRT)
    jr nz, nothit

    ; head = fruit 
    ld hl, (SCOREL)
    add l, 1
    daa l
    adc h, 0
    daa h
    ld (SCOREL), hl
    call DisplayScore

    ; update fruit position
pc1 ld w, 19
    call GetRandom
    ld (XFRT), a
    ld w, 16
    call GetRandom
    add a, 1
    ld (YFRT), a
    
    ld hl, (YFRT)
    call GetChar
    cp a, 0
    jr nz, pc1
    
    ld hl, (YFRT)
    ld w, 0
    ld a, FRUITCHAR
    call DrawChar

    jp slowdown


nothit

    ; move tail
    call qremove
    ld a, l

    ; check up
    bit PADB_UP, a
    jr z, tdown
    ld c, (YTAIL)
    cp c, 1
    jr nz, tu1
    ld c, 19
tu1 sub c, 1
    ld (YTAIL), c
    jp postmove

tdown
    ; check down
    bit PADB_DOWN, a
    jr z, tleft
    ld c, (YTAIL)
    cp c, 18
    jr nz, td1
    ld c, 0
td1 add c, 1
    ld (YTAIL), c
    jp postmove

tleft
    ; check left
    bit PADB_LEFT, a
    jr z, tright
    ld c, (XTAIL)
    cp c, 0
    jr nz, tl1
    ld c, 20
tl1 sub c, 1
    ld (XTAIL), c
    jp postmove

tright
    ; check right
    bit PADB_RIGHT, a
    jr z, postmove
    ld c, (XTAIL)
    cp c, 19
    jr nz, tr1
    ld c, 0
    jr tr2
tr1 add c, 1
tr2 ld (XTAIL), c
    
postmove    
    ; blank old tail position
    ld wa, 0
    ld hl, (OLDYTAIL)
    call DrawChar

    ; draw new tail
    ld xbc, TAILCONV
    ld xhl, 0
    call QPeek
    add xbc, xhl
    ld a, HEADBASE
    add a, (xbc)
    ld w, 2
    ld hl, (YTAIL)
    call DrawChar

slowdown
    ld w, (SPEED)
    call vbWait

; User shutdown? (Power off pressed?)

        cp      (rUSERS),0
        jp      z,Main      ; no

; Power off NGP

        ld      rw3,VECT_SHUTDOWN
        calr    SYSTEM_CALL

done:   jr      done


; vbWait
; Waits for n vertical blank interrupts
; inputs: w = number of VBIs to wait for

vbWait:
   push xwa
   ld (VBCOUNTER), 0
vbw1:
   ld a, (VBCOUNTER)
   cp a, w
   jr nz, vbw1
   pop xwa
   ret

WHICHHEAD db 0,1,2,0,3,0,0,0,4
DIRCONV   db 0,0,1,0,2,0,0,0,3
ODIRCONV  db 0,0,4,0,8,0,0,0,12
BODYTABLE db 6,0,9,8,0,6,10,7,7,8,5,0,10,9,0,5
TAILCONV  db 0,12,14,0,11,0,0,0,13

TXTSCORE  db 0,0,5,6,"SCORE"
TXTHIGH   db 13,0,2,6,"HI"  

DisplayValue
   div hl, 10
   ld e, h
   ld h, 0
   div hl, 10
   ld d, h
   ld c, l

   ld a, 40h
   add a, c
   ld (9800h), a

   ld a, 40h
   add a, d
   ld (9802h), a

   ld a, 40h
   add a, e
   ld (9804h), a
   ret

; 
; DrawChar
;
; Draws a character on the background plane
; inputs: wa = char, hl = xy
;
DrawChar
    push xhl
    push xde
    push xbc
    ld xde, 9800h
    ld c, h
    ld b, 0
    mul xbc, 2
    add de, bc
    ld h, 0
    mul xhl, 40h
    add de, hl
    ld (xde), a
    add xde, 1
    ld (xde), w
    pop xbc
    pop xde
    pop xhl
    ret
;
; DrawText
;
; XBC provides an address where the following is stored
; x, y, len, palette, "string"
;
DrawText
    push xhl
    push xde
    push xwa
    ld h, (XBC)
    add xbc, 1
    ld l, (XBC)
    add xbc, 1
    ld e, (XBC)
    add xbc, 1
    ld w, (XBC)
dt1
    add xbc, 1
    ld a, (XBC)
    cp a, ' '
    jr z, dt2
    sub a, 40h
    add a, TEXTBASE
    call DrawChar
dt2:
    add h, 1
    djnz e, dt1

    pop xwa
    pop xde
    pop xhl
    ret
    




; GetChar
;
; Gets the value of the character at the given xy position
; inputs: hl = xy
; outputs: wa = char
GetChar
    push xhl
    push xde
    push xbc
    ld xde, 9800h
    ld c, h
    ld b, 0
    mul xbc, 2
    add de, bc
    ld h, 0
    mul xhl, 40h
    add de, hl
    ld a, (xde)
    add xde, 1
    ld w, (xde)
    pop xbc
    pop xde
    pop xhl
    ret



VarInit
    ; initialise
    ld (SCOREH), 0
    ld (SCOREL), 0h
    ld (XHEAD), 10
    ld (YHEAD), 7
    ld (XTAIL), 10
    ld (YTAIL), 9
    ld l, PADF_UP
    ld (DIR), l
    call QADD
    call QADD

    ld (RANDOMOFS), 30
    ld w, 19
    call GetRandom
    ld (XFRT), a
    ld w, 16
    call GetRandom
    add a, 1
    ld (YFRT), a

    ld xbc, TXTSCORE
    call DrawText

    ld xbc, TXTHIGH
    call DrawText

    call DisplayScore

    ld hl, 1000h
    ld w, 6
    ld a, (HIH)
    call DisplayHex
    add h, 2
    ld a, (HIL)
    call DisplayHex

    ret

; initialise queue vars
QINIT
    LD (QHEAD), 0
    LD (QTAIL), 0
    RET

; adds L to queue
QADD
    PUSH XDE
    PUSH XHL
    LD XDE, QBASE
    ADD E, (QTAIL)
    LD (XDE), L
    LD L, (QTAIL)
    ADD L, 1
    LD (QTAIL), L
    POP XHL
    POP XDE
    RET

; removes L from QUEUE
QREMOVE
    PUSH XDE
    LD XDE, QBASE
    ADD E, (QHEAD)
    LD L, (XDE)
    LD E, (QHEAD)
    ADD E, 1
    LD (QHEAD), E
    POP XDE
    RET

; peeks at head of QUEUE and puts value in L
QPEEK
    PUSH XDE
    LD XDE, QBASE
    ADD E, (QHEAD)
    LD L, (XDE)
    POP XDE
    RET


; GetRandom
;
; Provides a random number in the range 0 - w
; Inputs: w upper range of random
; Outputs: a random number
GetRandom
    push xbc
    push xde
    ld xbc, BEGINRANDOM
    ld xde, 0
    ld e, (RANDOMOFS)
    add xbc, xde
    ld a, (xbc)
gr1 ld b, a
    sub a, w
    jr nc, gr1
    ld a, b

    ld c, (RANDOMOFS)
    add c, 1
    cp c, ENDRANDOM-BEGINRANDOM
    jr nz, gr2
    ld c, 0
gr2 ld (RANDOMOFS), c
    pop xde
    pop xbc
    ret

BEGINRANDOM
       db 124, 109, 224, 181, 223, 106, 62, 114, 114, 8, 66, 9, 173, 145
       db 17, 158, 39, 35, 127, 224, 128, 2, 40, 89, 202, 71, 177, 39
       db 230, 128, 242
ENDRANDOM


;
; DisplayHex
;
; Displays a hex value (e.g. FF) on the background plane
;
; inputs: w = palette, a = value; hl = xy
;
DisplayHex
    push xhl
    push xbc
    push xwa
    ld c, a
    and a, 0f0h
    srl 4, a
    add a, 40h
    call DrawChar
    ld a, c
    and a, 0fh
    add a, 40h
    add h, 1
    call DrawChar
    pop xwa
    pop xbc
    pop xhl
    ret

;
; DisplayScore
;
DisplayScore
    push xhl
    push xwa
    ld hl, 0600h
    ld w, 6
    ld a, (SCOREH)
    call DisplayHex
    add h, 2
    ld a, (SCOREL)
    call DisplayHex
    pop xwa
    pop xhl
    ret

;
; CheckDeath
;
CheckDeath
   ld hl, (YHEAD)
   call GetChar
   cp a, 0
   jr z, cd1
   
   ; we've hit something

   ; is it the fruit
   cp a, FRUITCHAR
   jr z, cd1

cd2
   ; we have eaten ourself - time to die 
   ld hl, (8308h) ; brown
   ld w, 020h
   Call vbwait
   ld (830eh), hl
   ld w, 020h
   Call vbwait
   ld (830ah), hl
   ld w, 020h
   Call vbwait
   ld (830ch), hl
   ld w, 020h
   Call vbwait

   ld a, 1
   ret

cd1
   ; clear A
   ld a, 0
   ret

;
; ValidControl
; inputs a = control
; outputs a = valid control
ValidControl
   and a, 15
   jr nz, vc1
   ld a, (DIR)
   ret
vc1
   ; first ensure no diagonals
   ld xbc, SGLCTRL
   ld xhl, 0
   ld l, a
   add xbc, xhl
   ld a, (xbc)
   
   ; now ensure not moving backwards
   ld xbc, BADCTRL
   ld xhl, 0
   ld l, a
   add xbc, xhl
   ld e, (xbc)
   cp e, (DIR)
   jr nz, vc2
   ld a, (DIR)

vc2 ; everything is fine
   ret

SGLCTRL db 1,1,2,1,4,1,2,1,8,1,2,1,4,1,2,1
BADCTRL db 0,2,1,0,8,0,0,0,4


;
; Clear the background plane
;
;
ClearScreen
   ld xbc, _SCR2RAM
   ld hl, 4C0h
cs1
   ld (xbc), 0
   add xbc, 1
   sub hl, 1
   jr nz, cs1
   ret


;
; Display the Title Screen and wait for key press
;
;
TXTSOFTWARE db 1,2,18,2,"NO FRILLS SOFTWARE"
TXTPRESENTS db 6,4,8,2,"PRESENTS"
TXTBUTTON   db 3,17,14,2,"PRESS BUTTON A"
TXTTITLE1   db 5,7,10,4,"lmpqtuxy|}"
TXTTITLE2   db 5,8,10,4,"norsvwz{~"
TXTME       db 1,12,18,2,"BY IVAN MACKINTOSH"
TXTSNAKE    db 4,10,12,2,"jbbbbbbbbbba"
TXTSPEED    db 5,15,5,2,"SPEED"
TXTSP1      db 11,15,4,2,"SLOW"
TXTSP2      db 11,15,4,2,"FAST"

TitleScreen
   call ClearScreen

   ld xbc, TXTSOFTWARE
   call DrawText
   ld xbc, TXTPRESENTS
   call DrawText
   ld xbc, TXTSNAKE
   call DrawText
   ld xbc, TXTTITLE1
   call DrawText
   ld xbc, TXTTITLE2
   call DrawText
   ld xbc, TXTME
   call DrawText
   ld xbc, TXTBUTTON
   call DrawText

   ; display correct speed setting
   ld xbc, TXTSPEED
   call DrawText

   ld xbc, TXTSP1
   ld a, (SPEED)
   cp a, 4
   jr nz, ts1
   ld xbc, TXTSP2
ts1:
   call DrawText

ts2:
   ; wait for button to be pressed or speed to be changed
   ld a, (rSL)
   bit PADB_LEFT, a
   jr z, ts3
   ld (SPEED), 8 ; slow speed
   ld xbc, TXTSP1
   call DrawText
ts3:
   bit PADB_RIGHT, a
   jr z, ts4
   ld (SPEED), 4 ; fast speed
   ld xbc, TXTSP2
   call DrawText
ts4:
   bit PADB_A, a
   jr z, ts2
ts5:
   ; now wait for button to be released
   ld a, (rSL)
   bit PADB_A, a
   jr nz, ts5
   ret
   
;
; UpdateHighScore
; 
; Is the current score larger than the hiscore
;
UpdateHighScore
   ld hl, (SCOREL)
   ld bc, (HIL)
   sub bc, hl
   jr nc, ups1

   ld (HIL), hl
ups1
   ret



VBlankInt:
    push xwa

    ; Update Watch Dog Timer to prevent CPU reset
    ld      (rWDCR),WD_CLR

    ; make the random number lookup table more random
    ld w, 25
    call GetRandom

    ; increment a counter
    inc (VBCOUNTER)
    pop xwa
    reti

SYSTEM_CALL:
        SystemCallCode
OS_VERSION:
        OsVersionCode


    include "charset.inc"   ; Snake + Font character data
   
    ; make the cartridge size correct
    org HEADER_OFS + 07FFFFh
    db 0FFh

